#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/resource.h>

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

static int zone = 1;

int verbose;

static int
get_status ()
{
  char msgbuf[1];
  i2c(zone, 2, 0, msgbuf, 1);
  return msgbuf[0];
}

static void
get_climate ()
{
  float t, h, f;
  char th[4];
  i2c(zone, 3, 0, th, 4);
  t = th[0] + th[1]/256.0;
  h = th[2] + th[3]/256.0;
  f = t * 9.0/5.0 + 32.0;
  printf("climate: %.2fC %.2fF %.2f%%\n", t, f, h);
}

int
main(int argc, char **argv)
{
  int s, os = -1;
  int i, ts, te;

  setbuf(stdout, 0);
  i2c_open ();

  zone = 1;

  for (i=1; i<argc; i++)
    {
      if (strcmp (argv[i], "-z") == 0)
	{
	  zone = atoi(argv[i+1]);
	  i ++;
	}
    }

  i = 0;
  ts = msclock();
  while (1)
    {
      s = get_status ();
      if (s & 0x80)
	{
	  te = msclock();
	  /*	  printf("%02x %8d %6d\n", s, te-ts, i);*/
	  i = 0;
	  ts = te;
	  get_climate();
	}
      else
	i++;
      if ((s & 0x0f) != os)
	{
	  int b;
	  int c = s ^ os;
	  for (b=0; b<4; b++)
	    if (c & (1<<b))
	      printf("button %d %s\n", b, s & (1<<b) ? "pressed" : "released");

	  os = s & 0x0f;
	}
      mspause(2);
    }

  i2c_close ();
  return 0;
}
