#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#include "i2c.h"
#include "gpio.h"

static int i2c_fd;
static struct i2c_msg msg[1];
static struct i2c_rdwr_ioctl_data mlist;
static int reg[6] = {-1,-1,-1,-1,-1,-1};

static int zone_addr[] = { 0, 10, 20, 30, 40, 50 };

int verbose;

static time_t last_good_i2c_time = 0;

static int zone_resets[] = { 64, 65, 66, 67, 68 };

static void
reset_all_r8cs(int reset)
{
  int i;
  for (i=0; i<5; i++)
    gpio_write (zone_resets[i], !reset);

  memset (reg, -1, sizeof(reg));
}

void
i2c_open()
{
  i2c_fd = open("/dev/i2c-0", O_RDWR);
  if (i2c_fd < 0)
    {
      perror ("/dev/i2c-0");
      exit (1);
    }

  time(&last_good_i2c_time);
  mlist.msgs = msg;
  mlist.nmsgs = 1;
}

void
i2c_close()
{
  close (i2c_fd);
}

int
i2c(int zone, int _reg, int do_write, void *buf, int len)
{
  time_t now;
  char msgbuf[2048];
  if (do_write)
    {
      memcpy (msgbuf+1, buf, len);
      msgbuf[0] = _reg;
      reg[zone] = _reg;
      msg[0].len = len+1;
      msg[0].buf = msgbuf;
    }
  else
    {
      if (_reg != reg[zone])
	if (i2c(zone, _reg, 1, 0, 0))
	  return 1;
      msg[0].len = len;
      msg[0].buf = buf;
    }
  msg[0].addr = zone_addr[zone];
  msg[0].flags = do_write ? 0 : I2C_M_RD;
  if (do_write && verbose)
    {
      int i;
      printf("w%d:", zone);
      for (i=0; i<msg[0].len; i++)
	printf(" %02x", msg[0].buf[i]);
      printf("\n");
    }
  if (ioctl (i2c_fd, I2C_RDWR, &mlist) < 0)
    {
      /*perror("ioctl:write");*/
      printf("\033[31m[z%d!]\033[0m", zone);
      time(&now);
      if ((int)(now - last_good_i2c_time) > 10)
	{
	  printf("\nresetting i2c and r8cs\n\n");
	  i2c_reset ();
	  time(&last_good_i2c_time);
	}
      return 1;
    }
  time(&last_good_i2c_time);
  if (!do_write && verbose)
    {
      int i;
      printf("r%d:", zone);
      for (i=0; i<msg[0].len; i++)
	printf(" %02x", msg[0].buf[i]);
      printf("\n");
    }
  /* The R8C's get confused if there isn't enough time between the
     stop and start states.  */
  uspause(50);
  return 0;
}

static void
i2c_probe()
{
  msg[0].len = 0;
  msg[0].buf = &msg;
  msg[0].addr = 0;
  msg[0].flags = 0;
  ioctl (i2c_fd, I2C_RDWR, &mlist);
}

void
i2c_reset()
{
  time_t now;
  struct tm *tm;

  time(&now);
  tm = localtime(&now);

  printf("%02d-%02d %02d:%02d:%02d i2c_reset\n",
	 tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);

  printf("r8c... ");
  reset_all_r8cs(1);
  i2c_close();
  sleep(2);
  printf("rmmod... ");
  system("rmmod i2c_pxa");
  sleep(2);
  printf("modprobe... ");
  system("modprobe i2c_pxa");
  sleep(2);
  printf("restart... ");
  i2c_open ();
  reset_all_r8cs(0);
  sleep(2);
  printf("probe...\n");
  i2c_probe();
}
