.include "r8c-ports.S"

I2C_TXOFF = 0x8d

#define CTRACE 0
//#define console_putc serial_putc
//#define IPTRACE	mov.w	#__LINE__,iptrace
#define IPTRACE

	.data

	.global	_status_byte
_status_byte:
	.byte	0x60

	.global	_i2c_state
_i2c_state:	
	.byte	0x60

;;; These are set from the command structs once we read a regno.
command_rx_start:
	.word	0
command_rx_byte:
	.word	0
command_tx_start:
	.word	0
command_tx_buf:
	.short	0
command_tx_count:
	.short	0

;;; These are the active values for writes in progress.
	.global	_i2c_tx_buffer
_i2c_tx_buffer:
	.short	0
	.global	_i2c_tx_count
_i2c_tx_count:
	.short	0

	.text
	.global	_i2c_init
_i2c_init:
	mov.b	#i2c_address,r1l
	rot.b	#1,r1l
	mov.b	r1l,sar		; slave address 0x5a, i2c format
	mov.b	#0x80,pmr	; port mode - i2c ports
	mov.b	#0x80,iccr1	; i2c enabled


	mov.b	#0x8d,iccr1	; i2c enabled, 100 kHz
	mov.b	#0x00,icmr	; 8 bits, msb first, no wait
	mov.b	#0xfc,icier	; all interrupts

	mov.b	#5,iic2aic	; enable interrupts

	mov.b	#0x7f,iccr2	; reset
	mov.b	#0x7d,iccr2	; reset

	mov.b	#0, icsr	; clear all flags

	rts

;;; This maps the value of the status byte to the handler we want.
irq_hash:
	.byte	28, 0, 4, 0, 8, 0, 4, 0,12,12,12,12,12,12,12,12 ; 00-0f
	.byte	24,24,24,24,24,24,24,24,12,12,12,12,12,12,12,12 ; 10-1f
	.byte	16,16, 0, 0,16,16, 0, 0,16,16, 0, 0,16,16, 0, 0 ; 20-2f
	.byte	16,16, 0, 0,16,16, 0, 0,16,16, 0, 0,16,16, 0, 0 ; 30-3f
	.byte	 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8 ; 40-4f
	.byte	 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8 ; 50-5f
	.byte	 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0 ; 60-6f
	.byte	 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0 ; 70-7f
	.byte	20,20, 4, 4,20,20, 4, 4,12,12, 4, 4,12,12, 4, 4 ; 80-8f
	.byte	24,24, 4, 4,24,24, 4, 4,24,24, 4, 4,24,24, 4, 4 ; 90-9f
	.byte	16,16, 0, 0,16,16, 0, 0,16,16, 0, 0,16,16, 0, 0 ; a0-af
	.byte	24,24, 0, 0,24,24, 0, 0,24,24, 0, 0,24,24, 0, 0 ; b0-bf
	.byte	 8, 8, 4, 4, 8, 8, 4, 4, 8, 8, 4, 4, 8, 8, 4, 4 ; c0-cf
	.byte	 8, 8, 4, 4, 8, 8, 4, 4, 8, 8, 4, 4, 8, 8, 4, 4 ; d0-df
	.byte	 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0 ; e0-ef
	.byte	 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0, 8, 8, 0, 0 ; f0-ff


	.data
handlers:
	.word	read_start	; 00 we just ack'd the address, RW was read
	.word	write_start	; 04 we just ack'd the address, RW was write
	.word	write_done	; 08 transmitter empty
	.word	i2c_reset	; 12 stop condition during read/write
	.global	_i2c_rx_handler
_i2c_rx_handler:
	.word	read_regno	; 16 byte has been read in
	.global	_i2c_tx_handler
_i2c_tx_handler:
	.word	write_again	; 20 transmit buffer empty
	.word	write_done	; 24 write failed
	.word	just_return	; 28 no interrupt (?)

	.global	_i2c_tx_start_handler
_i2c_tx_start_handler:
	.word	just_return

	;; Main entry - figure out what happened, and jump to the
	;; appropriate handler.
	.text
	.global	irq_i2c
irq_i2c:
	pushm	r0,r1,r2,r3,a0,a1
	fset	I
#if CTRACE
	mov.b	#'I',r0l
	jsr.w	console_putc
	mov.b	#'2',r0l
	jsr.w	console_putc
	stc	sp,r0
	jsr.w	console_print_r0
#endif
	mov.b	icsr,a0
	mov.b	a0,_i2c_state
#if 0
	cmp.b	#0x08,a0
	jeq	2f
	and.b	#0xfe,_p1_bits	; on
	mov.b	_p1_bits,p1
	jmp.b	1f
2:	
	or.b	#0x01,_p1_bits	; off
	mov.b	_p1_bits,p1
1:	
	cmp.b	#0,a0
#endif
	jeq	1f
	
#if CTRACE
	mov.b	a0,r0l
	jsr.w	console_putc
#endif
	mov.b	irq_hash[a0],a0
	IPTRACE
#if CTRACE
	push.w	a0
	mov.b	a0,r0l
	jsr.w	console_putc
	mov.w	handlers[a0],r0
	jsr.w	console_print_r0
	pop.w	a0
#endif
	jsri.a	handlers[a0]
1:	
	;; we can't do this, because we drop bytes if the handler
	;; takes too long.
;	mov.b	#0,icsr
#if CTRACE
	mov.b	icsr,r0l
	jsr.w	console_putc
#endif
;	mov.b	#0xee,r1l
;	jsr.w	_serial_putc
	popm	r0,r1,r2,r3,a0,a1
	reit
	
just_return:
	IPTRACE
	rts

i2c_reset:
	IPTRACE
	mov.b	#0, icsr	; clear all flags
	mov.b	#0x8d,iccr1	; i2c enabled, 100 kHz
	mov.b	#0xfc,icier	; all interrupts
	mov.b	#0x7f,iccr2	; reset
	mov.b	#0x7d,iccr2	; reset
;	bset	_iicrst,iccr2	; reset
;	bclr	_iicrst,iccr2	; reset release
	mov.b	#0, icsr	; clear all flags
	or.b	#0x01,_p1_bits	; off
	mov.b	_p1_bits,p1
	rts

#if CTRACE
	.global	_i2c_status
_i2c_status:	
	mov.b	#'R',r0l
	jsr.w	console_putc
	mov.b	sar,r0l
	jsr.w	console_putc
	mov.b	pmr,r0l
	jsr.w	console_putc
	mov.b	iccr1,r0l
	jsr.w	console_putc
	mov.b	icmr,r0l
	jsr.w	console_putc
	mov.b	#' ',r0l
	jsr.w	console_putc
	rts
#endif

read_start:
	IPTRACE
	mov.w	#read_regno,_i2c_rx_handler
	bclr	_aas,icsr
	mov.b	icdrr,r1l
	rts

read_byte:
	IPTRACE
	mov.b	icdrr,r1l	; read data byte
	mov.w	command_rx_byte,a0
	mov.w	#0,a1
;	jsr.w	_dumpregs
	jmpi.a	a1a0
	.byte	0
	
write_start:
	IPTRACE
	mov.b	icdrr,r1l
	bclr	_aas,icsr
#if CTRACE
	mov.b	#'W',r0l
	jsr.w	console_putc
#endif
	mov.w	command_tx_start,r1
	jz	1f
	mov.w	command_tx_start+2,r3
	jsri.a	r3r1
	jmp.b	2f
1:
	mov.w	command_tx_buf,_i2c_tx_buffer
#if CTRACE
	mov.b	command_tx_buf+1,r0l
	jsr.w	console_putc
	mov.b	command_tx_buf,r0l
	jsr.w	console_putc
#endif
	mov.w	command_tx_count,_i2c_tx_count
#if CTRACE
	mov.b	command_tx_count+1,r0l
	jsr.w	console_putc
	mov.b	command_tx_count,r0l
	jsr.w	console_putc
#endif
2:
	cmp.w	#0,_i2c_tx_count
	jeq	write_done

write_again:
	IPTRACE
#if CTRACE
	mov.b	#'w',r0l
	jsr.w	console_putc
#endif
	;; see if we have data to send
	cmp.w	#0,_i2c_tx_count
	jeq	write_again_last

	;; see if the buffer is empty
#if CTRACE
	mov.b	icsr,r0l
	jsr.w	console_putc
#endif
	tst.b	#0x80, icsr
	jz	write_again_more

	;; write one byte from the data to the tx buffer
#if CTRACE
	mov.b	#'>',r0l
	jsr.w	console_putc
#endif
	mov.w	_i2c_tx_buffer,a0
#if CTRACE
	mov.b	[a0],r0l
	jsr.w	console_putc
#endif
	mov.b	[a0],icdrt
	add.w	#1,_i2c_tx_buffer
	add.w	#-1,_i2c_tx_count

	jmp.b	write_again

write_again_more:
#if CTRACE
	mov.b	#'m',r0l
	jsr.w	console_putc
#endif
	bset	_tie,icier	; enable tx buffer empty interrupt
	rts
write_again_last:
#if CTRACE
	mov.b	#'l',r0l
	jsr.w	console_putc
#endif
	bclr	_tie,icier	; disable tx buffer empty interrupt
	bset	_teit,icier	; enable transmitter empty interrupt
	rts

write_done:
	IPTRACE
#if CTRACE
	mov.b	#'D',r0l
	jsr.w	console_putc
#endif
;	or.b	#0x01,_p1_bits
;	mov.b	_p1_bits,p1
	bclr	_tend,icsr
	bclr	_trs,iccr1
	mov.b	icdrr,r1l	; dummy read
	bclr	_tdre,icsr
	mov.b	#iccr2_sdaop, iccr2 ; generate stop condition
	bset	_tie,icier	; enable tx buffer empty interrupt
;	and.b	#0xfe,_p1_bits
;	mov.b	_p1_bits,p1
	rts

;;; ----------------------------------------------------------------------

empty_rx_start:
empty_rx_byte:
	mov.b	#'E',r0l
	jsr.w	console_putc
	mov.b	#'r',r0l
	jsr.w	console_putc
	mov.b	#'b',r0l
	jsr.w	console_putc
	rts

i2c_empty_cmd:
	.word	empty_rx_start
	.word	empty_rx_byte
	.word	0
	.short	0
	.short	0

	.data
i2c_command_structs:
	.short	i2c_debug_cmd
	.short	i2c_empty_cmd
	.short	i2c_status_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.short	i2c_empty_cmd
	.text

	.global	_i2c_register_command
_i2c_register_command:		; R1 = command, R2 = cmd
	mov.w	r1,a0
	shl.w	#1,a0
	mov.w	r2,i2c_command_structs[a0]
	rts

read_regno:
	IPTRACE
	mov.b	icdrr,a0	; read data byte
#if CTRACE
	mov.b	#'r',r0l
	jsr.w	console_putc
	mov.b	a0,r0l
	jsr.w	console_putc
#endif
	cmp.w	#11,a0
	jgeu	invalid_regno
	shl.w	#1,a0
	mov.w	i2c_command_structs[a0],a0
	mov.w	0[a0],command_rx_start
	mov.w	4[a0],command_rx_byte
	mov.w	8[a0],command_tx_start
	mov.w	12[a0],command_tx_buf
	mov.w	14[a0],command_tx_count

	mov.w	#read_byte,_i2c_rx_handler
	cmp.w	#0,command_rx_byte
	jnz	1f
	mov.w	#empty_rx_byte,command_rx_byte
1:	
	cmp.w	#0,command_rx_start
	jeq	1f
	jmpi.a	command_rx_start
	.byte	0
1:	rts

invalid_regno:
	push.w	a0
	jsr.w	console_prints
	.ascii	"invalid regno:	 \0"
	pop.w	a0
	mov.w	a0,r0
	jsr.w	console_print_r0
	jsr.w	console_prints
	.ascii	"\r\n\0"
	jmp.w	i2c_reset
	


;;; ----------------------------------------------------------------------

i2c_debug_cmd:
	.word	debug_rx_start
	.word	debug_rx_byte
	.word	0
	.short	debug_buffer
	.short	4

	.data
debug_buffer:
	.byte	0,0,0,0
debug_ptr:
	.byte	0

debug_rx_start:	
	mov.b	#0,debug_ptr
	rts

debug_rx_byte:
	mov.b	debug_ptr,a0
	mov.b	r1l,debug_buffer[a0]
	add.b	#1,a0
	and.b	#3,a0
	mov.b	a0,debug_ptr
	rts

;;; ----------------------------------------------------------------------

i2c_status_cmd:
	.word	0
	.word	0
	.word	0
	.short	_status_byte
	.short	1

;;; ----------------------------------------------------------------------
	.data
