.include "r8c-ports.S"

bufsize = 256
bufmask = 0xff

#define TRACE 0

	.bss
tx_buffer:
	.space	bufsize

	.global	tx_put
	.global	tx_get
	.global	count
	.data
tx_put:	.short	0
tx_get:	.short	0
high:	.short	bufsize/2
count:	.short	0

	.text
	.global	_serial_init
_serial_init:

	mov.b	#0x05,u0mr	; UART, 8N1
	mov.b	#0x00,u0c0	; LSB first, CMOS, buffered, f1
	mov.b	#0x01,u0c1	; enable transmitter
	mov.b	#0x30,ucon	; uart0 interrupt on buffer empty.
	mov.b	#0x81,u0brg	; 9600 baud at f1, 20MHz
;	mov.b	#0x00,u0brg	; 1.250M baud at f1, 20MHz
;	mov.b	#0x01,u0brg	; 625K baud at f1, 20MHz
;	mov.b	#0x07,u0brg	; 312K baud at f1, 20MHz

	mov.w	#6,r1
	mov.w	#i2c_serial_write_cmd,r2
	jsr.w	_i2c_register_command

	mov.w	#5,r1
	mov.w	#i2c_serial_mark_cmd,r2
	jsr.w	_i2c_register_command

	mov.w	#irq_uart0tx,irq_vector+0x44

	bset	3,pd3
	bclr	3,p3
	rts

;;; ----------------------------------------------------------------------

i2c_serial_write_cmd:
	.word	0
	.word	serial_write_rx_byte
	.word	0
	.short	0
	.short	0

i2c_serial_mark_cmd:
	.word	0
	.word	serial_mark_rx_byte
	.word	0
	.short	0
	.short	0

serial_write_rx_byte:
	mov.b	#0,r1h
	jmp.w	serial_enqueue

serial_mark_rx_byte:
	mov.b	#0,r1h
	jmp.w	serial_set_space


;;; ----------------------------------------------------------------------

;;; pass the byte in r1l.  ONLY works if there is no other serial traffic,
;;; so that the interrupts never happen.
	.global	_serial_putc
_serial_putc:
	btst	1,u0c1
	jz	_serial_putc
	mov.b	r1l,u0tb
	rts
	.global	serial_putc
serial_putc:
	btst	1,u0c1
	jz	serial_putc
	mov.b	r0l,u0tb
	rts

;;; ----------------------------------------------------------------------

;;; pass the byte in r1l
	.global	serial_enqueue
serial_enqueue:
;	mov.b	#0x71,p1
	pushm	r0,a0
	
#if TRACE
	mov.b	#'E',r0l
	jsr.w	console_putc
	mov.b	r1l,r0l
	jsr.w	console_putc
#endif
	
	mov.w	tx_put,a0
	mov.b	r1l,tx_buffer[a0]
	inc.w	a0
	and.w	#bufmask,a0
	mov.w	a0,tx_put

	add.w	#1,count
	jsr.w	serial_check_space

;	mov.b	count,r0l
;	jsr.w	console_putc

	cmp.w	#1,count
	jne	1f
	jsr.w	serial_start_transmitter

1:
;	mov.b	#0x72,p1
	popm	r0,a0
	rts

serial_start_transmitter:
;	mov.b	#0x73,p1
#if TRACE
	mov.b	u0c1,r0l
	jsr.w	console_putc
#endif
	btst	1, u0c1		; see if the buffer is empty
	jz	sst_buffer_is_full ; it's full, just enqueue
	jsr.w	serial_dequeue_to_buffer

	;; if the buffer is still empty, it's because the byte was
	;; immediately moved to the transmitter.  The next enqueue
	;; will immediately fill the buffer and enable interrupts.
#if TRACE
	mov.b	u0c1,r0l
	jsr.w	console_putc
#endif
	btst	1, u0c1		; see if the buffer is empty
	jnz	sst_done	; yes, we're done

sst_buffer_is_full:
	;; the transmitter is busy, so enable the interrupt handlers.
	mov.b	#0x30,ucon	; uart0 interrupt on buffer empty.
	mov.b	#5,s0tic
	
	mov.w	#2,r1
	jsr.w	_led_on

#if TRACE
	mov.b	#'I',r0l
	jsr.w	console_putc
#endif

sst_done:	
	rts

;;; pass new low-water margin in r1
	.global	serial_set_space
serial_set_space:
	mov.w	#bufsize, r0
	sub.w	r1,r0
	mov.w	r0,high
;	mov.b	#0xcc, p1
;	mov.b	r0h, p1
	;; fall through

serial_check_space:
	mov.b	_status_byte, r1l
	mov.w	count,r0
	cmp.w	high,r0
	bmle	6,r1
	cmp.w	#0,r0		; txbuf empty
	bmz	c
	band	3,u0c0		; and transmitter empty
	band	1,u0c1		; and buffer empty
	bmc	5,r1		; means we're not transmitting
	mov.b	r1l,_status_byte
	rts

serial_dequeue_to_buffer:
	mov.w	tx_get,a0
#if TRACE
	mov.b	tx_buffer[a0],r0l
	jsr.w	console_putc
#endif
	mov.b	tx_buffer[a0],u0tb
	inc.w	a0
	and.w	#bufmask,a0
	mov.w	a0,tx_get
	add.w	#-1,count
	rts

irq_uart0tx:
;	fset	I
	pushm	r0,r1,r2,r3,a0,a1

#if TRACE
	mov.b	#'X',r0l
	jsr.w	console_putc
#endif

	cmp.w	#0,count
	jnz	more_to_xmit

	btst	0,ucon
	jz	wait_for_tx_done

	mov.b	#0,s0tic	; buffer empty, too - set status
#if TRACE
	mov.b	#'i',r0l
	jsr.w	console_putc
#endif
	mov.w	#2,r1
	jsr.w	_led_off
	jmp.b	irq_common

wait_for_tx_done:	
	bset	0,ucon		; now we'll wait for tx done
	jmp.b	irq_common
	
more_to_xmit:
	jsr.w	serial_dequeue_to_buffer
	
irq_common:
	jsr.w	serial_check_space
	popm	r0,r1,r2,r3,a0,a1
	reit

	.global	_serial_ping
_serial_ping:
	pushm	r0,r1,r2,r3,a0,a1

	fclr	I
	cmp.w	#0,count
	jeq	_serial_ping_done ; no data, OK
	btst	1,u0c1
	jz	_serial_ping_done ; transmitter busy, OK

	;; at this point, we have data, but we're not transmitting it.

#if 1
	jsr.w	console_prints
	.ascii	"\r\nkicking serial queue\r\n\0"
#endif
	jsr.w	serial_dequeue_to_buffer

_serial_ping_done:
	fset	I
	popm	r0,r1,r2,r3,a0,a1
	rts
