#include <stdarg.h>

#include "console.h"
#include "r8c-ports.h"
#include "globals.h"
#include "i2c.h"
#include "leds.h"

#define BUFSIZE 256

static byte tx_buffer[BUFSIZE];
static int tx_put = 0;
static int tx_get = 0;
static int high = BUFSIZE/2;
static volatile int count = 0;

static void
serial_check_space()
{
  if (count < high)
    status_byte |= 0x40;
  else
    status_byte &= ~ 0x40;
  if (count == 0)
    status_byte |= 0x20;
  else
    status_byte &= ~ 0x20;
  /*printf("status byte set to %b at %w %w\n", status_byte, tx_get, tx_put);*/
}

static void
serial_mark_rx (unsigned char b)
{
  high = BUFSIZE - b;
  /*printf("mark set to %d out of %d\n", high, BUFSIZE);*/
  serial_check_space ();
}

static void
serial_write_rx (unsigned char b)
{
  if (count >= BUFSIZE)
    return;
  tx_buffer[tx_put] = b;
  tx_put = (tx_put+1) % BUFSIZE;
  count ++;
  led_on(YELLOW);
  serial_check_space ();
}

void
serial_send()
{
  while (count > 0)
    {
      while (u0c1.ti == 0)
	;
      u0tb = tx_buffer[tx_get];
      tx_get = (tx_get+1) % BUFSIZE;
      asm("fclr I");
      count --;
      asm("fset I");
      serial_check_space ();
    }
  led_off(YELLOW);
}

/*----------------------------------------------------------------------*/

static const i2c_setup_struct serial_write_cmd = {
  0, serial_write_rx, 0, 0, 0
};

static const i2c_setup_struct serial_mark_cmd = {
  0, serial_mark_rx, 0, 0, 0
};

serial_init()
{
  u0mr.b = 5;
  u0c0.b = 0;
  u0c1.b = 1;
  ucon.b = 0x30;
  u0brg = 0x81; /* 9600 baud */

  i2c_register_command (5, &serial_mark_cmd);
  i2c_register_command (6, &serial_write_cmd);
}
